/*
 * atn - auto telnet
 *
 * 1995-02-19 Written by Kouichi Hirabayashi
 * 1995-03-04 New syntax.
 * 2018-05-06 modified for Freebsd clang
 *
 * Feel free to modify, copy, use and distribute but absolutely no waranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>


#define MAILCMD	"/usr/bin/mail"		/* mail command */
#define LOGFILE	"log"			/* default log file */
#define HOST	"localhost"		/* default telnet host */
#define TIMEOUT	30			/* default l_time value */
#define LNDELAY	"100000L"		/* default line delay */
#define SERIAL	"/dev/cuaa0"	/* default serial port */

struct salist {	/* string/action list */
  char *pat;	/* search string */
  char *act;	/* action */
  char *ptr;	/* pointer to search string */
  struct salist *next;
};

struct salist *sahead;
int nsalist;

void usage();
int _auto();
int flush();
void Abort(int);
int cmdOpen();
int Logf();
int execute(char *);
int loop(char *);
int option(char *);
int _speed(char *);
int parse(char *);
int lex(char *);
int getch();
int ungetch(int);
int mscan(struct salist *, char *);
int scan(char *);
struct salist *addlist(struct salist *, char *, char *);
int freelist(struct salist *);
#ifdef DEBUG
void prtlist(struct salist *);
#endif
void synerr(char *, int);
void error(char *, char *);
int action(char *);
void onalarm();
void Abort(int);
int sendf(char *);
void delay();
void _sleep(int);
int comBrk();
int mail();
int etos(char *, char *);
int output(int);
int flush();
int serOpen(char *, int, int);
int serClose(int);
int etos(char *, char *);

/* result of action */
#define QUIT		-1
#define CONTINUE	0
#define BREAK		1

/* Kanji code */
#define EUC		0
#define SJIS		1


FILE *ifp;		/* script file pointer */
char *isp;		/* script pointer (used if ifp == NULL) */

int pid;	/* telnet pid */

int ifd;	/* telnet input */
int ofd;	/* telnet output */
int code;	/* code system (default is EUC) */
int lfd;	/* log file */
int mflg;	/* mail on/off */
unsigned int lndelay;	/* line delay in uSec */
int l_time = TIMEOUT;
int l_retry = 1;	/* loop counter */
int nofile;	/* missing "&s"file" */
FILE *mfp;	/* mail file */

int print;	/* &p{ .. } */

/* script option */
char host[128]; 	/* telnet host */
char port[128];		/* port # */
char serial[128];	/* serial device */
int speed;		/* speed */
int cflag = CS8;	/* c_cflag */
char command[128];	/* command */
char logfile[128];	/* log file */
char mailadr[128];	/* mail address */
int lineno = 1;		/* lino no. */
#ifndef OLDSIGNAL
struct sigaction saalrm, osaalrm;
#endif

/* command line option */
char *optmadr;
char *optlog;

int _sflg;	/* serial */
int vflg;	/* verbose */

char *cmd;
char buf[BUFSIZ];

int main(argc, argv) char **argv;
{
  char c, *s;
  FILE *efopen();

  cmd = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-')
	for (s = argv[0]+1; *s; s++)
		switch (*s) {
		case 'm': case 'l':
			c = *s;
			if (s[1])
				s++;
			else {
				if (argc < 2)
					usage();
				s = *++argv; --argc;
			}
			if (c == 'm')
				optmadr = s;
			else
				optlog = s;
			s += strlen(s) - 1;
			break;
		case 'v':
			vflg++;
			break;
		default:
			usage();
		}
  if (argc == 0)
	ifp = stdin;
  else if (argc == 1)
	ifp = efopen(*argv, "r");
  else
	usage();
#ifdef OLDSIGNAL
  signal(SIGALRM, onalarm);
#else
  saalrm.sa_handler = onalarm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
#endif
  _auto(ifp);
  flush();
  if (mfp != NULL)
	pclose(mfp);
  sleep(1);

  Abort(0);
}

void usage()
{
  fprintf(stderr, "Usage: %s [-m mail_address] [-l logfile] script_file\n",
	cmd);
  exit(1);
}

int cmdOpen()
{
  int i, in[2], out[2];
  static char *s, *t, *av[50], str[256];

  if (pipe(in) == -1)
	error("pipe() faild", (char *)0);
  if (pipe(out) == -1)
	error("pipe() failed", (char *)0);
  if ((pid = fork()) == -1)
	error("fork() failed", (char *)0);
  else if (pid == 0) {
	dup2(out[0], 0);
	close(out[0]);
	close(out[1]);
	dup2(in[1], 1);
	dup2(1, 2);
	close(in[0]);
	close(in[1]);
/*printf("*comamnd(%s)*\n", command);*/
	strcpy(str, command);
	for (i = 0, s = str; *s; ) {
		while (*s && (*s == ' ' || *s == '\t'))
			s++;
		if (*s == '\0')
			break;
		t = s;
		av[i++] = s;
		while (*t && *t != ' ' && *t != '\t')
			t++;
		if (*t)
			*t++ = '\0';
		s = t;
	}
	av[i++] = NULL;
	execvp(av[0], av);
/*	execlp("sh", "sh", "-c", command, (char *)0);*/
	error("execlp() failed", (char *)0);
  }
  else {
	ofd = out[1];
	close(out[0]);
	ifd = in[0];
	close(in[1]);
  }
  return 0;
}

int _auto(fp) FILE *fp;
{
  int c, r;
  char tmps[128], *s, *strcpy();
  long atol();

  do {
	if ((s = fgets(buf, BUFSIZ, fp)) == NULL)
		return -1;
	lineno++;
  } while (s && (*s == '\n' || *s == '#' || option(s)));
  for (s = buf+strlen(buf)-1; buf <= s; s--)
	ungetc(*s, fp);

/*
printf("command(%s)\n", command);
*/
  if (*logfile == '\0')
	strcpy(logfile, LOGFILE);
  if (*host == '\0')
	strcpy(host, HOST);
  if (optmadr != NULL && *optmadr != '\0') {
	strcpy(mailadr, optmadr);
	mail();
  }
  if (optlog != NULL) {
	if (*optlog != '\0')
		strcpy(logfile, optlog);
	Logf();
  }
  if (!_sflg && *command == '\0') {
	if (*port == '\0')
		sprintf(command, "telnet %s", host);
	else
		sprintf(command, "telnet %s %s", host, port);
  }

  if (_sflg) {
	if (*serial == '\0')
		strcpy(serial, SERIAL);
	if (speed == 0)
		speed = B9600;
	if (cflag == 0)
		cflag = CS8;
	ifd = ofd = serOpen(serial, speed, cflag);
  }
  else
	cmdOpen();

  while (parse(buf) != EOF)
	if (execute(buf) == QUIT)
		break;

  if (_sflg)
	serClose(ifd);
  else {
	c = '\004';	/* EOT */
	write(ofd, &c, 1);	/* send EOT to telnet command */
  }

  return 0;
}

int option(s) char *s;
{
  int f, r = 0;
  char c, *t, *u, tmps[128], *strchr();
  long atol();

  t = NULL;
  while (*s) {
	f = 0;
	t = NULL;
	if (strncmp(s, "host", 4) == 0)
		t = host;
	else if (strncmp(s, "port", 4) == 0)
		t = port;
	else if (strncmp(s, "log", 3) == 0)
		t = logfile;
	else if (strncmp(s, "mail", 4) == 0)
		t = mailadr;
	else if (strncmp(s, "command", 7) == 0)
		t = command;
	else if (strncmp(s, "serial", 6) == 0) {
		_sflg++;
		t = serial;
	}
	else if (strncmp(s, "delay", 5) == 0) {
		f = 1;
		t = tmps;
	}
	else if (strncmp(s, "code", 4) == 0) {
		f = 2;
		t = tmps;
	}
	else if (strncmp(s, "speed", 5) == 0) {
		f = 3;
		t = tmps;
	}
	else if (strncmp(s, "parity", 6) == 0) {
		f = 4;
		t = tmps;
	}
	else
		break;
	while (*s && strchr(": \t", *s) == NULL)
		s++;	/* skip keyword */
	do {
		while (*s && strchr(": \t", *s) != NULL)
			s++;	/* skip white */
		while (*s && strchr(" ,#\t\n", *s) == NULL)
			*t++ = *s++;	/* copy keyword */
		if (strchr(" \t", *s) != NULL)
			*t++ = ' ';
	} while (*s && strchr(",#\n", *s) == NULL);
	*t = '\0';
	if (t != NULL) {
		r = 1;
		if (f == 1)
			lndelay = atol(tmps);
		else if (f == 2) {
			if (strcmp(tmps, "SJIS") == 0)
				code = SJIS;
			else
				code = EUC;
		}
		else if (f == 3)
			speed = _speed(tmps);
		else if (f == 4) {
			if (strcmp(tmps, "even") == 0) {
				cflag &= ~CS8;
				cflag |= CS7|PARENB;
			}
			else if (strcmp(tmps, "odd") == 0) {
				cflag &= ~CS8;
				cflag |= CS7|PARENB|PARODD;
			}
		}
	}
	while (*s && strchr(",; \t\n", *s) != NULL) {
		if (*s == '#') {
			while (*s && *s != '\n')
				s++;
			lineno++;
		}
		else
			s++;	/* skip delimiter */
	}
  }
  return r;
}

int _speed(s) char *s;
{
  static struct {
	char *s;
	int n;
  } t[] = {
	"300", B300,
	"1200", B1200,
	"2400", B2400,
	"4800", B4800,
	"9600", B9600,
	"19200", B19200,
	"38400", B38400,
	"115200", B115200,
	NULL, B9600
  };
  int i;

  for (i = 0; t[i].s != NULL; i++)
	if (strcmp(s, t[i].s) == 0)
		break;
  if (t[i].s == NULL)
	fprintf(stderr, "Warning: illegal serial speed: set to 9600 bps.\n");
/*printf("speed(%s)(%d)\n", s, t[i].n);*/
  return t[i].n;
}

int execute(src) char *src;
{
  int c, r = 0;
  char *p, *s, *t, *ssp, *strsave();
  FILE *sfp;

  p = s = strsave(src);
  if ((c = *s) == '{') {
	++s;	/* skip '{' */
	t = s + strlen(s) - 1;
	c = *t; *t = '\0';
	if (code == SJIS)
		etos(s, s);
	r = scan(s);
	*t = c;
	print = 0;
  }
  else if (c == '[') {
	r = loop(s);
  }
  else if (c == '(') {
	if (nofile) {
		t = s + strlen(s) - 1;
		c = *t; *t = '\0';
		sfp = ifp; ssp = isp; ifp = NULL; isp = s + 1;
		while (parse(buf) != EOF)
			if ((r = execute(buf)) != CONTINUE)
				break;
		*t = c;
		ifp = sfp; isp = ssp;
	}
  }
  else {
	r = action(s);
  }
  free(p);
  return r;
}

int loop(s) char *s;
{
  char *pat;
  char *act;
  char *xact;
  struct salist *head, *addlist();
  char *t;
  int c, d, r = 0;
  int sretry, stime;
  FILE *sfp;
  char *ssp, *strsave(), *strcpy();

/*
printf("loop(%s)(%d)(%d)\n", s, l_retry, l_time);
*/
  sfp = ifp; ssp = isp; ifp = NULL; isp = ++s;	/* switch source, skip '[' */
  sretry = l_retry; stime = l_time;	/* set time & counter to local */

  head = (struct salist *)NULL;
  xact = NULL;
  do {
	if ((c = parse(buf)) == EOF)
		synerr("Unexpected EOF", 0);
	if (c == '{' || c == '?') {
		if (c == '{') {
			t = buf + strlen(buf) - 1;
			d = *t; *t = '\0';
			if (code == SJIS)
				etos(buf+1, buf+1);
			pat = strsave(buf+1);
			*t = d;
		}
		d = c;
		t = buf;
		do {
			if ((c = parse(t)) == EOF)
				synerr("Unexpected EOF", 0);
			if (c != ',' && c != ']')
				t += strlen(t);
		} while (c != ',' && c != ']');
		*t = '\0';
		if (d == '{') {
			if (*buf == '\0')
				strcpy(buf, "B");	/* default action */
			act = strsave(buf);
			head = addlist(head, pat, act);
		}
		else {
			if (xact == NULL)
				xact = strsave(buf);
		}
	}
	else
		synerr("missing {} or ?", 0);
  } while (c != ']');
#ifdef DEBUG
  prtlist(head);
#endif
  r = mscan(head, xact);

  /* free list, xaction */
  freelist(head);
  if (xact != NULL)
	free(xact);

  if (r > 0)
	--r;
/*
printf("loop_exit(%d)\n", r);
*/
  l_retry = sretry; l_time = stime;
  ifp = sfp; isp = ssp;
  return r;
}

int parse(s) char *s;
{
  int c, d, e, n, n1, n2;
  char *t;

  if ((c = lex(s)) == EOF)
	return EOF;
/*printf("**lex(%s)**\n", s);*/
  if (c == '[' || c == '(') {
	d = c;
	e = (c == '[') ? ']' : ')';
	*s++ = c;
	n = 1;
	t = s;
	n1 = n2 = 0;
	if (c == '[')
		++n1;
	else if (c == '(')
		++n2;
	do {
		c = lex(t);
		t += strlen(t);
		if (c == d)
			++n;
		else if (c == e)
			--n;
		if (c == '[')
			++n1;
		else if (c == '(')
			++n2;
		else if (c == ']')
			--n1;
		else if (c == ')')
			--n2;
/*printf("loop(%s)c(%c)n(%d)\n", text, c, n);*/
	} while (c != EOF && (c != e || n > 0));
	if (c == EOF)
		synerr("unexpected EOF", 0);
	if (n1 > 0)
		synerr("missing ']'", 0);
	if (n2 > 0)
		synerr("missing ')'", 0);
	*t = '\0';
	return d;
  }
  else
	return c;
  return c;
}

int lex(s) char *s;
{
  int c, d, e, i, j;

  do {	/* skip leading while space */
	if ((c = getch()) == EOF)
		return EOF;
  } while (strchr(" \t\n", c) != NULL);
  while (c == '#') {	/* skip comment */
	do {
		c = getch();
	} while (c != EOF && c != '\n');
	if (c == EOF)
		return EOF;
	lineno++;
	if ((c = getch()) == EOF)
		return EOF;
  }
  if (strchr("[](),?", c) != NULL) {	/* token */
	*s++ = c;
	*s = '\0';
	return c;	/* token */
  }
  else if (c == '{') {	/* pattern */
	d = c;
	e = '}';
	*s++ = c;
	do {
		c = getch();
		if (c == '\\') {
			c = getch();
			if (c == 'r')
				c = '\r';
			else if (c == 'n')
				c = '\n';
			else if (c == 't')
				c = '\t';
			else if (c == 'b')
				c = '\b';
			else if (c == 'f')
				c = '\f';
			else if ('0' <= c && c <= '9') {
				j = c - '0';
				for (i = 1; i < 3; i++) {
					c = getch();
					j = j * 8 + c - '0';
				}
				c = j;
			}
			*s++ = c; c = 'x';	/* dummy */
		}
		else if (c & 0x80) {
			*s++ = c; *s++ = getch();
		}
		else
			*s++ = c;
	} while (c != EOF && c != e);
	*s = '\0';
	if (c == EOF)
		synerr("unexpected EOF: missing %c", d);
	return d;
  }
  else {	/* action */
	do {
		*s++ = c;
		if (c == '"') {
			e = c;
			do {
				c = getch();
				*s++ = c;
				if (c == '\\') {
					c = getch();
					*s++ = c;
					c = 'x';	/* dummy */
				}
			} while (c != EOF && c != e);
		}
		c = getch();
	} while (c != EOF && strchr("[]{}(),?# \t\n", c) == NULL);
	if (c != EOF)
		ungetch(c);
	*s = '\0';
	return 'a';	/* action */
  }
  return 0;
}

int getch()
{
  int c;

  if (ifp != NULL)
	c = fgetc(ifp);
  else {
	if (isp == NULL || *isp == '\0')
		c = EOF;
	else
		c = *isp++;
  }
  return c;
}

int ungetch(c)
{
  if (c == EOF)
	fprintf(stderr, "Warning: unget(EOF)\n");
  if (ifp != NULL)
	ungetc(c, ifp);
  else
	*--isp = c;
  return 0;
}


int mscan(head, xact) struct salist *head; char *xact;
{
  struct salist *p;
  int i, r;
  char c;
  FILE *sfp;
  char *ssp;

/*
printf("mscan(%s)\n", xact);
*/
  for (i = 0; i < l_retry || l_retry == 0; i++) {
	for (p = head; p != NULL; p = p->next)
		p->ptr = p->pat;	/* initialize */
	alarm(l_time);
	for (;;) {
		c = '\0';
		do {
			if ((r = read(ifd, &c, 1)) < 0)
				goto g1;	/* alarm */
/*
if (c >= ' ') printf("(%c)",c);
else printf("(%02x)", c & 0xff);
fflush(stdout);
*/
		} while (c == '\0');
		output(c);
		for (p = head; p != NULL; p = p->next) {
/*printf("(%s)(%c)\n", p->ptr, c);*/
			if (c != *p->ptr++) {	/* !match (redo from start) */
				p->ptr = p->pat;
				if (c == *p->pat)
					p->ptr++;
			}
			else if (*p->ptr == '\0') {	/* match */
				alarm(0);	/* match */
				sfp = ifp; ssp = isp; sfp = NULL; isp = p->act;
				while (parse(buf) != EOF) {
					if ((r = execute(buf)) != CONTINUE)
						break;
				}
				ifp = sfp; isp = ssp;
				if (r == CONTINUE) {	/* 0 */
					i--;
					goto g2;	/* next */
				}
				else
					return r;	/* BREAK, QUIT */
			}
		}
	}
g1:
	if (xact != NULL && *xact != '\0') {
		sfp = ifp; ssp = isp; sfp = NULL; isp = xact;
		while (parse(buf) != EOF) {
			if ((r = execute(buf)) != CONTINUE)
				break;
		}
		ifp = sfp; isp = ssp;
		if (r == CONTINUE) {	/* 0 */
			i--;
			goto g2;	/* next */
		}
		else
			return r;	/* BREAK, QUIT */
	}
	else
		return QUIT;	/* failed */
g2:	;
  }
  return QUIT;	/* failed */
}

int scan(str) char *str;
{
  char c, *t;
  int i;

  alarm(l_time);
  for (t = str; *t; ) {
	c = '\0';
	do {
		if (read(ifd, &c, 1) < 0) {
			printf("ALARM.\n");
			return QUIT;	/* alarm */
	}
/*
if (c >= ' ') printf("(%c)",c);
else printf("(%02x)", c & 0xff);
fflush(stdout);
*/
	} while (c == '\0');
/*
if (c >= ' ') printf("(%c)(%s)",c, t);
else printf("(%02x)", c & 0xff);
fflush(stdout);
*/
	if (print)
		write(1, &c, 1);
	output(c);
	if (c != *t++) {
		t = str;
		if (c == *str)
			t++;
	}
  }
  alarm(0);
  return CONTINUE;	/* match */
}

struct salist *addlist(head, pat, act) struct salist *head; char *pat, *act;
{
  struct salist *p, *q;
  char *emalloc();

  p = (struct salist *)emalloc(sizeof(struct salist));
  p->pat = pat;
  p->act = act;
  p->ptr = p->pat;
  p->next = NULL;
  if (head == NULL)
	head = p;
  else {
	for (q = head; ; q = q->next)
		if (q->next == NULL) {
			q->next = p;
			break;
		}
  }
  return head;
}

int freelist(head) struct salist *head;
{
  struct salist *p, *q;

  if (head != NULL) {
	for (p = head; ; p = q) {
		q = p->next;
		if (p->pat != NULL)
			free(p->pat);
		if (p->act != NULL)
			free(p->act);
		free(p);
		if (q == NULL)
			break;
	}
  }
  head = NULL;
  return 0;
}

#ifdef DEBUG
prtlist(head) struct salist *head;
{
  struct salist *p;

  if (head != NULL) {
	for (p = head; p != NULL; p = p->next) {
		printf("salist: pat(%s),act(%s)\n", p->pat, p->act);
	}
  }
}
#endif

char *emalloc(n)
{
  char *p;

  if ((p = malloc(n)) == NULL)
	error("memory over", (char *)0);
  return p;
}

char *strsave(s) char *s;
{
  char *p, *emalloc(), *strcpy();

  p = emalloc(strlen(s)+1);
  return strcpy(p, s);
}

FILE * efopen(file, mode) char *file, *mode;
{
  FILE *fp;

  if ((fp = fopen(file, mode)) == NULL)
	error("can't open %s", file);
  return fp;
}

void synerr(s, t) char *s;
{
  fprintf(stderr, "syntax error: line(%d): ", lineno);
  fprintf(stderr, s, t);
  fprintf(stderr, "\n");
  exit(1);
}

void error(s, t) char *s, *t;
{
  fprintf(stderr, s, t);
  fprintf(stderr, "\n");
  exit(1);
}

int action(s) char *s;
{
  int fd, i, j;
  char c, d, *t;

  if (strcmp(s, "&p") == 0) {
	print = 1;
	return 0;
  }
  while ((c = *s++) != '\0') {
	switch (c) {
	case 'S': case 's':
		c = 's';
		goto g1;
	case 'L': case 'l':
		c = 'l';
		goto g1;
	case '&': c = *s++;
g1:
		if (c == 's') {
			d = *s++;
			for (t = s; *s && *s != d; )
				s++;
			d = *s; *s = '\0';
			sendf(t);
			if ((*s = d) != '\0')
				s++;
		}
		else if (c == 'l')
			Logf();
		else if (c == 'm')
			mail();
		else if (c == 'p') {
			d = *s++;
			for (t = s; *s && *s != d; )
				s++;
			d = *s; *s = '\0';
			fprintf(stdout, "\n%s\n", t);
			*s = d;
			if ((*s = d) != '\0')
				s++;
		}
		else
			fprintf(stderr, "Illegal command: &%c\n", c);
		break;
	case '`':
	case '"':
	case '\'': while ((d = *s++) && d != c && d) {
			if (d == '^') {
				if ((d = *s++) != '^')
					d &= 037;	/* CTRL code */
			}
			else if (d == '\\') {
				if ((d = *s++) == 'n')
					d = '\n';
				else if (d == 't')
					d = '\t';
				else if (d == 'v')
					d = '\v';
				else if (d == 'b')
					d = '\b';
				else if (d == 'r')
					d = '\r';
				else if ('0' <= d && d <= '7') {
					for (--s, i = j = 0; i < 3; s++, i++) {
						d = *s;
						if (d < '0' || '7' < d)
							break;
						j = j * 8 + d - '0';
					}
					d = j;
				}
				else {
					fprintf(stderr,
	"Illegal special charcter sequence: <\\%c>\n", d);
				}
			}
			write(ofd, &d, 1);
		}
		break;
	case 'B': case 'b':
		for (j = 0; '0' <= *s && *s <= '9'; )
			j = j * 10 + *s++ - '0';
		if (j == 0)
			j = 1;	/* default */
		return j;	/* break */
	case 'C': case 'c': comBrk(ofd);
		break;
	case 'D': case 'd':
		write(ofd, "\177", 1);	/* DEL */
		break;
	case 'E': case 'e':
		write(ofd, "\r", 1);
		break;
	case 'P': case 'p':
		fprintf(stderr, "\n   >>> pause (press `RETURN`) <<<\n");
		do {
			read(0, &d, 1);
		} while (d != '\n');
		break;
	case 'Q': case 'q':
		return QUIT;
		break;
	case 'N': case 'n':
	case 'T': case 't':
	case 'W': case 'w':
		i = 0;
		for (t = s; *s >= '0' && *s <= '9'; )
			s++;
		d = *s;
		*s = '\0';
		i = atoi(t);
		if ( c == 'T' || c == 't')
			l_time = i;
		else if (c == 'N' || c == 'n')
			l_retry = i;
		else
			_sleep(i);
		*s = d;
		break;
	case ' ':
	case ',':
	case '\t':
	case '\n':
		break;
	case '#':
		while (*s != '\n')
			s++;
		break;
	default:
		fprintf(stderr,
			"action syntax error: >>%c<< %s\n", c, s);
		break;
	}
  }
  return CONTINUE;
}

void onalarm()
{
#ifdef OLDSIGNAL
  signal(SIGALRM, onalarm);
#else
  saalrm.sa_handler = onalarm;
  sigemptyset(&saalrm.sa_mask);
  saalrm.sa_flags = 0;
  sigaction(SIGALRM, &saalrm, &osaalrm);
#endif
}

void Abort(r)
{
  char c = '\004';

  if (_sflg)
	serClose(ifd);
  else {
	write(ofd, &c, 1);
	sleep(2);
	if (pid > 0) {
printf("kill %d SIGTERM\n", pid);
		kill(pid, SIGTERM);
	}
  }
  exit(r);
}

int sendf(file) char *file;
{
  FILE *fp;
  int n;
  char c, *s, *t, tmps[BUFSIZ];

  if ((fp = fopen(file, "r")) == NULL) {
	nofile = 1;
	return -1;
  }
  else
	nofile = 0;
  while (fgets(tmps, BUFSIZ, fp) != NULL) {
	if (code == SJIS)
		etos(tmps, tmps);
	if (strncmp(tmps, "{{", 2) == 0) {
		s = tmps + 2;
		t = tmps + strlen(tmps) - 3;
		if (strncmp(t, "}}\n", 3) == 0) {
			c = *t; *t = '\0';
			if (scan(s) == QUIT)
				Abort(1);
			*t = c;
			continue;
		}
	}
	n = strlen(tmps);
	*(tmps + n - 1) = '\r';
	write(ofd, tmps, n);
#ifdef EATECHO
	do {	/* eat echo */
		read(ifd, &c, 1);
		output(c);
	} while (c != '\n');
#endif
	delay();
  }
  return fclose(fp);
}

void delay()
{
#ifndef NODELAY
  usleep(lndelay);
#endif
}

void _sleep(n)
{
  if (n == 0)
	usleep(100000L);
  else
	sleep(n);
}

int comBrk()
{
  char c = '\003';

  return write(ofd, &c, 1);
}

int Logf()
{
  char *strcpy();

  if (lfd == 0) {
	if ((lfd = open(logfile, O_CREAT|O_WRONLY|O_APPEND, 0666)) == -1)
		error("can't open %s", logfile);
  }
  else {
	close(lfd);
	lfd = 0;
  }
  return 0;
}

int mail()
{
  char tmps[256], *strcpy();

  if (mfp == NULL) {
	if (*mailadr != '\0') {
		sprintf(tmps, "%s %s", MAILCMD, mailadr);
		if ((mfp = popen(tmps, "w")) == NULL)
			error("popen(%s) failed", tmps);
		mflg = 1;
	}
	else
		fprintf(stderr, "Warning: mail address is not specified.\n");
  }
  else
	mflg = mflg ? 0 : 1;
  return 0;
}

/* code conversion */

int etos(s, t) char *s, *t;
{
  int hi;
  int c;
  char *u, *v, tmps[BUFSIZ];

  for (u = tmps; (*u++ = *s++) != '\0'; )
	;
  hi = 0;
  for (u = tmps; *u; ) {
	c = *u++ & 0xff;
	if (hi) {
		if (hi != 0x0e) {	/* !KANA */
			c &= 0x7f;
			c += (hi & 1) ? 0x1f : 0x7d;
			if (c >= 0x7f)
				c++;
			hi = ((hi - 0x21) >> 1) + 0x81;
			if (hi > 0x9f)
				hi += 0x40;
			*t++ = hi;
		}
		*t++ = c;
		hi = 0;
	}
	else if (c & 0x80)	/* KANJI 1 */
		hi = c & 0x7f;
	else	/* AN */
		*t++ = c;
  }
  *t = '\0';
  return 0;
}

static char lbuf[BUFSIZ];
static char *lbufp = lbuf;

int output(c) int c;
{
  char ch;

  *lbufp++ = c;
/*
ch = c;
write(1, &ch, 1);
*/
  if (c == '\n' || lbufp == lbuf+BUFSIZ) {
	if (vflg)
		write(1, lbuf, lbufp - lbuf);
	if (lfd != 0)
		write(lfd, lbuf, lbufp - lbuf);
	if (mfp != NULL && mflg)
		fwrite(lbuf, sizeof(char), lbufp - lbuf, mfp);
	lbufp = lbuf;
  }
  return 0;
}

int flush()
{
  if (vflg)
	write(1, lbuf, lbufp - lbuf);
  if (lfd != 0)
	write(lfd, lbuf, lbufp - lbuf);
  if (mfp != NULL && mflg)
	fwrite(lbuf, sizeof(char), lbufp - lbuf, mfp);
  lbufp = lbuf;
#if 0
  if (mfp)
	fwrite(&c, sizeof(char), 1, mfp);	/* send EOT to mail command */
#endif
  return 0;
}

int serOpen(port, speed, cflag) char *port;
{
  int fd;
  struct termios tios;

  /* reopen to enabme c_cc[VMIN] and c_cc[VTIME] */
  if ((fd = open(port, O_RDWR)) == -1)
	error("can't open %s", port);
  tios.c_iflag = 0;
  tios.c_oflag = 0;
  tios.c_cflag = CLOCAL|CREAD|HUPCL|CRTSCTS|cflag;
  tios.c_lflag = 0;
#if 0
  tios.c_cc[VMIN] = 0;
  tios.c_cc[VTIME] = 10;
#else
  tios.c_cc[VMIN] = 1;
  tios.c_cc[VTIME] = 0;
#endif
  if (cfsetospeed(&tios, speed) == -1)
	error("cfsetospeed() failed.", (char *)0);
  if (cfsetispeed(&tios, speed) == -1)
	error("cfsetispeed() failed.", (char *)0);
  if (tcsetattr(fd, TCSANOW, &tios) == -1)
	error("ioctl(TCSANOW) failed", (char *)0);

  return fd;
}

int serClose(fd)
{
  return close(fd);
}
